@@ -1,11 +1,10 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2011-2018  Ruby-GNOME2 Project Team
- *  Copyright (C) 2002-2004  Ruby-GNOME2 Project Team
+ *  Copyright (C) 2002-2023  Ruby-GNOME Project Team
  *  Copyright (C) 2002-2003  Masahiro Sakai
- *  Copyright (C) 1998-2000 Yukihiro Matsumoto,
- *                          Daisuke Kanda,
- *                          Hiroshi Igarashi
+ *  Copyright (C) 1998-2000  Yukihiro Matsumoto,
+ *                           Daisuke Kanda,
+ *                           Hiroshi Igarashi
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -31,6 +30,18 @@ VALUE RG_TARGET_NAMESPACE;
 static VALUE eNoPropertyError;
 static GQuark RUBY_GOBJECT_OBJ_KEY;
 
+VALUE
+rbg_cGLibObject(void)
+{
+    return RG_TARGET_NAMESPACE;
+}
+
+gboolean
+rbg_is_object(VALUE object)
+{
+    return RVAL2CBOOL(rb_obj_is_kind_of(object, RG_TARGET_NAMESPACE));
+}
+
 /* deperecated */
 void
 rbgobj_add_abstract_but_create_instance_class(G_GNUC_UNUSED GType gtype)
@@ -52,7 +63,9 @@ weak_notify(gpointer data, G_GNUC_UNUSED GObject *where_the_object_was)
 }
 
 static void
-holder_relatives_mark(gpointer key, gpointer value, gpointer user_data)
+holder_relatives_mark(G_GNUC_UNUSED gpointer key,
+                      gpointer value,
+                      G_GNUC_UNUSED gpointer user_data)
 {
     VALUE rb_relative = (VALUE)value;
     rb_gc_mark(rb_relative);
@@ -98,11 +111,10 @@ static const rb_data_type_t rg_glib_object_type = {
     {
         holder_mark,
         holder_free,
-        NULL,
     },
     NULL,
     NULL,
-    RUBY_TYPED_FREE_IMMEDIATELY,
+    RUBY_TYPED_FREE_IMMEDIATELY | RUBY_TYPED_FROZEN_SHAREABLE,
 };
 
 void
@@ -120,6 +132,15 @@ rbgobj_object_add_relative(VALUE rb_gobject, VALUE rb_relative)
     }
 }
 
+void
+rbgobj_gobject_remove_relative(GObject *gobject, VALUE rb_relative)
+{
+    gobj_holder *holder = g_object_get_qdata(gobject, RUBY_GOBJECT_OBJ_KEY);
+    if (holder && holder->rb_relatives) {
+        g_hash_table_remove(holder->rb_relatives, (gpointer)(rb_relative));
+    }
+}
+
 void
 rbgobj_object_remove_relative(VALUE rb_gobject, VALUE rb_relative)
 {
@@ -129,13 +150,12 @@ rbgobj_object_remove_relative(VALUE rb_gobject, VALUE rb_relative)
                          &rg_glib_object_type,
                          holder);
     if (holder->rb_relatives) {
-        g_hash_table_remove(holder->rb_relatives,
-                            (gpointer)(rb_relative));
+        g_hash_table_remove(holder->rb_relatives, (gpointer)(rb_relative));
     }
 }
 
 static gboolean
-rbgobj_object_remove_relatives_body(gpointer key,
+rbgobj_object_remove_relatives_body(G_GNUC_UNUSED gpointer key,
                                     gpointer value,
                                     gpointer user_data)
 {
@@ -305,6 +325,12 @@ rg_s_new_bang(int argc, VALUE *argv, VALUE self)
     return result;
 }
 
+static VALUE
+rg_s_init(int argc, VALUE *argv, VALUE self)
+{
+    return RUBY_Qnil;
+}
+
 struct param_setup_arg {
     GObjectClass* gclass;
     GParameter* params;
@@ -314,8 +340,14 @@ struct param_setup_arg {
 };
 
 static VALUE
-_params_setup(VALUE arg, struct param_setup_arg *param_setup_arg)
+_params_setup(VALUE arg,
+              VALUE rb_param_setup_arg,
+              G_GNUC_UNUSED int argc,
+              G_GNUC_UNUSED const VALUE *argv,
+              G_GNUC_UNUSED VALUE block)
 {
+    struct param_setup_arg *param_setup_arg =
+        (struct param_setup_arg *)rb_param_setup_arg;
     guint index;
     VALUE name, val;
     GParamSpec* pspec;
@@ -349,16 +381,20 @@ _params_setup(VALUE arg, struct param_setup_arg *param_setup_arg)
 }
 
 static VALUE
-gobj_new_body(struct param_setup_arg* arg)
+gobj_new_body(VALUE rb_arg)
 {
-    rb_iterate(rb_each, (VALUE)arg->params_hash, _params_setup, (VALUE)arg);
+    struct param_setup_arg *arg = (struct param_setup_arg *)rb_arg;
+    ID id_each;
+    CONST_ID(id_each, "each");
+    rb_block_call(arg->params_hash, id_each, 0, NULL, _params_setup, (VALUE)arg);
     return (VALUE)g_object_newv(G_TYPE_FROM_CLASS(arg->gclass),
                                 arg->param_size, arg->params);
 }
 
 static VALUE
-gobj_new_ensure(struct param_setup_arg* arg)
+gobj_new_ensure(VALUE rb_arg)
 {
+    struct param_setup_arg *arg = (struct param_setup_arg *)rb_arg;
     guint i;
     g_type_class_unref(arg->gclass);
     for (i = 0; i < arg->param_size; i++) {
@@ -381,10 +417,10 @@ rbgobj_gobject_new(GType gtype, VALUE params_hash)
     if (NIL_P(params_hash)) {
         result = g_object_newv(gtype, 0, NULL);
     } else {
-        size_t param_size;
+        guint param_size;
         struct param_setup_arg arg;
 
-        param_size = NUM2INT(rb_funcall(params_hash, rb_intern("length"), 0));
+        param_size = NUM2UINT(rb_funcall(params_hash, rb_intern("length"), 0));
 
         arg.param_size = param_size;
         arg.gclass = G_OBJECT_CLASS(g_type_class_ref(gtype));
@@ -481,140 +517,146 @@ gobj_s_properties(int argc, VALUE* argv, VALUE self)
     return ary;
 }
 
-static VALUE type_to_prop_setter_table;
-static VALUE type_to_prop_getter_table;
+static GHashTable *rbg_type_to_prop_setter_tables;
+static GMutex rbg_type_to_prop_setter_tables_mutex;
+static GHashTable *rbg_type_to_prop_getter_tables;
+static GMutex rbg_type_to_prop_getter_tables_mutex;
 
-void
-rbgobj_register_property_setter(GType gtype, const char *name, RValueToGValueFunc func)
+static void
+rbg_register_property_accessor(GHashTable *tables,
+                               GMutex *mutex,
+                               GType gtype,
+                               const char *name,
+                               gpointer accessor)
 {
-    GObjectClass* oclass;
-    GParamSpec* pspec;
+    g_mutex_lock(mutex);
 
-    VALUE table = rb_hash_aref(type_to_prop_setter_table, INT2FIX(gtype));
-    if (NIL_P(table)){
-        table = rb_hash_new();
-        rb_hash_aset(type_to_prop_setter_table, INT2FIX(gtype), table);
+    GHashTable *table = g_hash_table_lookup(tables, GUINT_TO_POINTER(gtype));
+    if (!table) {
+        table = g_hash_table_new(g_str_hash, g_str_equal);
+        g_hash_table_insert(tables, GUINT_TO_POINTER(gtype), table);
     }
 
-    oclass = g_type_class_ref(gtype);
-    pspec = g_object_class_find_property(oclass, name);
+    GObjectClass *gclass = g_type_class_ref(gtype);
+    GParamSpec *pspec = g_object_class_find_property(gclass, name);
+    g_hash_table_insert(table,
+                        (gchar *)g_param_spec_get_name(pspec),
+                        accessor);
+    g_type_class_unref(gclass);
 
-    rb_hash_aset(table, CSTR2RVAL(g_param_spec_get_name(pspec)),
-                 Data_Wrap_Struct(rb_cData, NULL, NULL, func));
+    g_mutex_unlock(mutex);
+}
 
-    g_type_class_unref(oclass);
+static gpointer
+rbg_get_property_accessor(GHashTable *tables,
+                          GMutex *mutex,
+                          GType gtype,
+                          const char *name)
+{
+    g_mutex_lock(mutex);
+
+    gpointer accessor = NULL;
+    GHashTable *table = g_hash_table_lookup(tables, GUINT_TO_POINTER(gtype));
+    if (table) {
+        GObjectClass *gclass = g_type_class_ref(gtype);
+        GParamSpec *pspec = g_object_class_find_property(gclass, name);
+        accessor = g_hash_table_lookup(table, g_param_spec_get_name(pspec));
+        g_type_class_unref(gclass);
+    }
+
+    g_mutex_unlock(mutex);
+
+    return accessor;
+}
+
+void
+rbgobj_register_property_setter(GType gtype,
+                                const char *name,
+                                RValueToGValueFunc func)
+{
+    rbg_register_property_accessor(rbg_type_to_prop_setter_tables,
+                                   &rbg_type_to_prop_setter_tables_mutex,
+                                   gtype,
+                                   name,
+                                   func);
 }
 
 void
 rbgobj_register_property_getter(GType gtype, const char *name, GValueToRValueFunc func)
 {
-    GObjectClass* oclass;
-    GParamSpec* pspec;
-
-    VALUE table = rb_hash_aref(type_to_prop_getter_table, INT2FIX(gtype));
-    if (NIL_P(table)){
-        table = rb_hash_new();
-        rb_hash_aset(type_to_prop_getter_table, INT2FIX(gtype), table);
-    }
-
-    oclass = g_type_class_ref(gtype);
-    pspec = g_object_class_find_property(oclass, name);
-
-    rb_hash_aset(table, CSTR2RVAL(g_param_spec_get_name(pspec)),
-                 Data_Wrap_Struct(rb_cData, NULL, NULL, func));
-
-    g_type_class_unref(oclass);
+    rbg_register_property_accessor(rbg_type_to_prop_getter_tables,
+                                   &rbg_type_to_prop_getter_tables_mutex,
+                                   gtype,
+                                   name,
+                                   func);
 }
 
 static VALUE
 rg_set_property(VALUE self, VALUE prop_name, VALUE val)
 {
-    GParamSpec* pspec;
     const char* name;
-
     if (SYMBOL_P(prop_name))
         name = rb_id2name(SYM2ID(prop_name));
     else
         name = StringValuePtr(prop_name);
 
-    pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(RVAL2GOBJ(self)),
-                                         name);
-
+    GParamSpec* pspec =
+        g_object_class_find_property(G_OBJECT_GET_CLASS(RVAL2GOBJ(self)),
+                                     name);
     if (!pspec)
         rb_raise(eNoPropertyError, "No such property: %s", name);
-    else {
-        // FIXME: use rb_ensure to call g_value_unset()
-        RValueToGValueFunc setter = NULL;
-        GValue gval = G_VALUE_INIT;
 
-        g_value_init(&gval, G_PARAM_SPEC_VALUE_TYPE(pspec));
-
-        {
-            VALUE table = rb_hash_aref(type_to_prop_setter_table,
-                                       INT2FIX(pspec->owner_type));
-            if (!NIL_P(table)){
-                VALUE obj = rb_hash_aref(table, CSTR2RVAL(g_param_spec_get_name(pspec)));
-                if (!NIL_P(obj))
-                    Data_Get_Struct(obj, void, setter);
-            }
-        }
-        if (setter) {
-            setter(val, &gval);
-        } else {
-            rbgobj_rvalue_to_gvalue(val, &gval);
-        }
-
-        g_object_set_property(RVAL2GOBJ(self), name, &gval);
-        g_value_unset(&gval);
-
-        G_CHILD_SET(self, rb_intern(name), val);
-
-        return self;
+    RValueToGValueFunc setter =
+        rbg_get_property_accessor(rbg_type_to_prop_setter_tables,
+                                  &rbg_type_to_prop_setter_tables_mutex,
+                                  pspec->owner_type,
+                                  name);
+    // FIXME: use rb_ensure to call g_value_unset()
+    GValue gval = G_VALUE_INIT;
+    g_value_init(&gval, G_PARAM_SPEC_VALUE_TYPE(pspec));
+    if (setter) {
+        setter(val, &gval);
+    } else {
+        rbgobj_rvalue_to_gvalue(val, &gval);
     }
+    g_object_set_property(RVAL2GOBJ(self), name, &gval);
+    g_value_unset(&gval);
+
+    G_CHILD_SET(self, rb_intern(name), val);
+
+    return self;
 }
 
 static VALUE
 rg_get_property(VALUE self, VALUE prop_name)
 {
-    GParamSpec* pspec;
     const char* name;
-
     if (SYMBOL_P(prop_name))
         name = rb_id2name(SYM2ID(prop_name));
     else
         name = StringValuePtr(prop_name);
 
-    pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(RVAL2GOBJ(self)),
-                                         name);
-
+    GParamSpec *pspec =
+        g_object_class_find_property(G_OBJECT_GET_CLASS(RVAL2GOBJ(self)),
+                                     name);
     if (!pspec)
         rb_raise(eNoPropertyError, "No such property: %s", name);
-    else {
-        // FIXME: use rb_ensure to call g_value_unset()
-        GValueToRValueFunc getter = NULL;
-        GValue gval = G_VALUE_INIT;
-        VALUE ret;
 
-        {
-            VALUE table = rb_hash_aref(type_to_prop_getter_table,
-                                       INT2FIX(pspec->owner_type));
-            if (!NIL_P(table)){
-                VALUE obj = rb_hash_aref(table, CSTR2RVAL(g_param_spec_get_name(pspec)));
-                if (!NIL_P(obj))
-                    Data_Get_Struct(obj, void, getter);
-            }
-        }
+    GValueToRValueFunc getter =
+        rbg_get_property_accessor(rbg_type_to_prop_getter_tables,
+                                  &rbg_type_to_prop_getter_tables_mutex,
+                                  pspec->owner_type,
+                                  name);
+    // FIXME: use rb_ensure to call g_value_unset()
+    GValue gval = G_VALUE_INIT;
+    g_value_init(&gval, G_PARAM_SPEC_VALUE_TYPE(pspec));
+    g_object_get_property(RVAL2GOBJ(self), name, &gval);
+    VALUE ret = getter ? getter(&gval) : GVAL2RVAL(&gval);
+    g_value_unset(&gval);
 
-        g_value_init(&gval, G_PARAM_SPEC_VALUE_TYPE(pspec));
-        g_object_get_property(RVAL2GOBJ(self), name, &gval);
+    G_CHILD_SET(self, rb_intern(name), ret);
 
-        ret = getter ? getter(&gval) : GVAL2RVAL(&gval);
-        g_value_unset(&gval);
-
-        G_CHILD_SET(self, rb_intern(name), ret);
-
-        return ret;
-    }
+    return ret;
 }
 
 static VALUE rg_thaw_notify(VALUE self);
@@ -699,7 +741,6 @@ rg_type_name(VALUE self)
     return CSTR2RVAL(G_OBJECT_TYPE_NAME(RVAL2GOBJ(self)));
 }
 
-#if GLIB_CHECK_VERSION(2, 26, 0)
 typedef struct {
     VALUE transform_from_callback;
     VALUE transform_to_callback;
@@ -828,7 +869,6 @@ rg_bind_property(gint argc, VALUE *argv, VALUE self)
 
     return rb_binding;
 }
-#endif
 
 static VALUE
 rg_initialize(int argc, VALUE *argv, VALUE self)
@@ -852,6 +892,15 @@ rg_initialize(int argc, VALUE *argv, VALUE self)
     gobj = rbgobj_gobject_new(RVAL2GTYPE(self), params_hash);
 
     G_INITIALIZE(self, gobj);
+
+    rb_funcall(self, rb_intern("initialize_post"), 0);
+
+    return Qnil;
+}
+
+static VALUE
+rg_initialize_post(VALUE self)
+{
     return Qnil;
 }
 
@@ -936,6 +985,9 @@ rbgobj_class_init_func(gpointer g_class, G_GNUC_UNUSED gpointer class_data)
 
     g_object_class->set_property = set_prop_func;
     g_object_class->get_property = get_prop_func;
+
+    VALUE rb_class = GTYPE2CLASS(G_TYPE_FROM_CLASS(g_class));
+    rb_funcall(rb_class, rb_intern("init"), 0);
 }
 
 void
@@ -1000,6 +1052,11 @@ rbgobj_register_type(VALUE klass, VALUE type_name, GClassInitFunc class_init)
                                             StringValueCStr(type_name),
                                             info,
                                             0);
+        if (type == G_TYPE_INVALID) {
+            rb_raise(rb_eArgError,
+                     "failed to register type: <%s>",
+                     StringValueCStr(type_name));
+        }
 
         rbgobj_register_class(klass, type, TRUE, TRUE);
 
@@ -1042,11 +1099,11 @@ rg_s_type_register(int argc, VALUE *argv, VALUE self)
 void
 Init_gobject_gobject(void)
 {
-    RG_TARGET_NAMESPACE = G_DEF_CLASS_WITH_GC_FUNC(G_TYPE_OBJECT, "Object", mGLib,
+    RG_TARGET_NAMESPACE = G_DEF_CLASS_WITH_GC_FUNC(G_TYPE_OBJECT, "Object", rbg_mGLib(),
                                                   gobj_mark, NULL);
 
 #ifdef G_TYPE_INITIALLY_UNOWNED
-    G_DEF_CLASS(G_TYPE_INITIALLY_UNOWNED, "InitiallyUnowned", mGLib);
+    G_DEF_CLASS(G_TYPE_INITIALLY_UNOWNED, "InitiallyUnowned", rbg_mGLib());
 #endif
 
     RUBY_GOBJECT_OBJ_KEY = g_quark_from_static_string("__ruby_gobject_object__");
@@ -1054,6 +1111,8 @@ Init_gobject_gobject(void)
     rb_define_alloc_func(RG_TARGET_NAMESPACE, rbgobj_object_alloc_func);
     RG_DEF_SMETHOD_BANG(new, -1);
 
+    RG_DEF_SMETHOD(init, 0);
+
     rbg_define_singleton_method(RG_TARGET_NAMESPACE, "property", &gobj_s_property, 1);
     rbg_define_singleton_method(RG_TARGET_NAMESPACE, "properties", &gobj_s_properties, -1);
     RG_DEF_SMETHOD(install_property, -1);
@@ -1069,24 +1128,31 @@ Init_gobject_gobject(void)
     RG_DEF_METHOD_P(destroyed, 0);
 
     RG_DEF_METHOD(initialize, -1);
+    RG_DEF_PRIVATE_METHOD(initialize_post, 0);
     rbg_define_method(RG_TARGET_NAMESPACE, "ref_count", gobj_ref_count, 0); /* for debugging */
     RG_DEF_METHOD_P(floating, 0); /* for debugging */
     RG_DEF_METHOD(unref, 0);
     RG_DEF_METHOD(inspect, 0);
     RG_DEF_METHOD(type_name, 0);
 
-#if GLIB_CHECK_VERSION(2, 26, 0)
     RG_DEF_METHOD(bind_property, -1);
-    G_DEF_CLASS(G_TYPE_BINDING_FLAGS, "BindingFlags", mGLib);
-#endif
+    G_DEF_CLASS(G_TYPE_BINDING_FLAGS, "BindingFlags", rbg_mGLib());
 
-    eNoPropertyError = rb_define_class_under(mGLib, "NoPropertyError",
+    eNoPropertyError = rb_define_class_under(rbg_mGLib(), "NoPropertyError",
                                              rb_eNameError);
 
-    rb_global_variable(&type_to_prop_setter_table);
-    rb_global_variable(&type_to_prop_getter_table);
-    type_to_prop_setter_table = rb_hash_new();
-    type_to_prop_getter_table = rb_hash_new();
+    rbg_type_to_prop_setter_tables =
+        g_hash_table_new_full(g_direct_hash,
+                              g_direct_equal,
+                              NULL,
+                              (GDestroyNotify)g_hash_table_unref);
+    g_mutex_init(&rbg_type_to_prop_setter_tables_mutex);
+    rbg_type_to_prop_getter_tables =
+        g_hash_table_new_full(g_direct_hash,
+                              g_direct_equal,
+                              NULL,
+                              (GDestroyNotify)g_hash_table_unref);
+    g_mutex_init(&rbg_type_to_prop_getter_tables_mutex);
 
     /* subclass */
     RG_DEF_SMETHOD(type_register, -1);
