@@ -1,6 +1,6 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2002-2017  Ruby-GNOME2 Project Team
+ *  Copyright (C) 2002-2021  Ruby-GNOME Project Team
  *  Copyright (C) 2002,2003  Masahiro Sakai
  *
  *  This library is free software; you can redistribute it and/or
@@ -31,9 +31,10 @@ static VALUE rb_cMutex;
 #endif
 static VALUE lookup_class_mutex;
 
+static ID id_gtype;
+static ID id_lock;
 static ID id_new;
 static ID id_superclass;
-static ID id_lock;
 static ID id_unlock;
 static GHashTable *gtype_to_cinfo;
 static VALUE klass_to_cinfo;
@@ -48,133 +49,105 @@ typedef struct {
 } RGObjClassInfoDynamic;
 
 typedef struct {
-    VALUE parent;
     GType gtype;
-    gboolean create_class;
-} RGObjClassByGtypeData;
+    const gchar *name;
+    VALUE module;
+    VALUE parent;
+} RGObjDefineClassData;
 
 static void
-cinfo_mark(RGObjClassInfo* cinfo)
+cinfo_mark(void *data)
 {
+    RGObjClassInfo *cinfo = data;
     rb_gc_mark(cinfo->klass);
 }
 
-const RGObjClassInfo *
-rbgobj_lookup_class(VALUE klass)
+static void
+cinfo_free(void *data)
 {
-    VALUE data = rb_hash_aref(klass_to_cinfo, klass);
-    if (!NIL_P(data)){
-        RGObjClassInfo* cinfo;
-        Data_Get_Struct(data, RGObjClassInfo, cinfo);
-        return cinfo;
-    }
-
-    if (TYPE(klass) == T_CLASS) {
-        VALUE super;
-        super = rb_funcall(klass, id_superclass, 0);
-        return rbgobj_lookup_class(super);
-    }
-
-    rb_raise(rb_eRuntimeError, "can't get gobject class information");
+    RGObjClassInfo *cinfo = data;
+    g_hash_table_remove(gtype_to_cinfo, GUINT_TO_POINTER(cinfo->gtype));
+    xfree(cinfo->name);
+    xfree(cinfo->data_type);
+    xfree(cinfo);
 }
 
-static const RGObjClassInfo *rbgobj_lookup_class_by_gtype_without_lock(GType gtype,
-                                                                       VALUE parent,
-                                                                       gboolean create_class);
+static RGObjClassInfo *
+rbgobj_class_info_define_without_lock(GType gtype,
+                                      const gchar *name,
+                                      VALUE module,
+                                      VALUE parent);
 
-static VALUE
-get_superclass(GType gtype)
+static void
+rbgobj_class_info_fill_name(RGObjClassInfo *cinfo)
 {
-    VALUE super_class;
+    VALUE rb_name;
 
-    if (rbgobj_convert_get_superclass(gtype, &super_class))
-        return super_class;
-
-    switch (gtype) {
-      case G_TYPE_PARAM:
-      case G_TYPE_OBJECT:
-        return cInstantiatable;
-      case G_TYPE_BOXED:
-        return rb_cObject;
-      case G_TYPE_POINTER:
-        return rb_cData;
-      case G_TYPE_ENUM:
-      case G_TYPE_FLAGS:
-        return rb_cObject;
-      default:
-      {
-          GType parent_type;
-
-          parent_type = g_type_parent(gtype);
-          if (parent_type == G_TYPE_INVALID) {
-              return cInstantiatable;
-          } else {
-              const RGObjClassInfo *cinfo_super;
-              cinfo_super =
-                  rbgobj_lookup_class_by_gtype_without_lock(parent_type,
-                                                            Qnil,
-                                                            TRUE);
-              return cinfo_super->klass;
-          }
-      }
+    if (cinfo->name) {
+        return;
     }
+
+    if (!RB_TYPE_P(cinfo->klass, RUBY_T_CLASS)) {
+        return;
+    }
+
+    rb_name = rb_funcall(cinfo->klass, rb_intern("name"), 0);
+    if (NIL_P(rb_name)) {
+        return;
+    }
+
+    cinfo->name = RB_ALLOC_N(char, RSTRING_LEN(rb_name) + 1);
+    memcpy(cinfo->name, RSTRING_PTR(rb_name), RSTRING_LEN(rb_name));
+    cinfo->name[RSTRING_LEN(rb_name)] = '\0';
+    cinfo->data_type->wrap_struct_name = cinfo->name;
 }
 
-static const RGObjClassInfo *
-rbgobj_lookup_class_by_gtype_without_lock(GType gtype, VALUE parent,
-                                          gboolean create_class)
+static rb_data_type_t *
+rbgobj_class_info_create_data_type(VALUE klass)
 {
+    rb_data_type_t *data_type;
+
+    data_type = RB_ZALLOC(rb_data_type_t);
+    data_type->wrap_struct_name = "RGObjClassInfo";
+    data_type->function.dmark = cinfo_mark;
+    data_type->function.dfree = cinfo_free;
+    if (RB_TYPE_P(klass, RUBY_T_CLASS) && klass != rb_cObject) {
+        VALUE p = RCLASS_SUPER(klass);
+        while (p != rb_cObject) {
+            if (RB_TYPE_P(p, RUBY_T_DATA) && RTYPEDDATA_P(p)) {
+                data_type->parent = RTYPEDDATA_TYPE(p);
+                break;
+            }
+            p = RCLASS_SUPER(p);
+        }
+    }
+    data_type->flags = RUBY_TYPED_FREE_IMMEDIATELY;
+
+    return data_type;
+}
+
+static RGObjClassInfo *
+rbgobj_class_info_register_without_lock(GType gtype, VALUE klass)
+{
+    rb_data_type_t *data_type;
+    RGObjClassInfo *cinfo;
     GType fundamental_type;
-    RGObjClassInfo* cinfo;
-    RGObjClassInfoDynamic* cinfod;
-    void* gclass = NULL;
+    RGObjClassInfoDynamic *cinfod;
+    void *gclass = NULL;
     VALUE c;
 
-    if (gtype == G_TYPE_INVALID)
-        return NULL;
-
-    cinfo = g_hash_table_lookup(gtype_to_cinfo, GUINT_TO_POINTER(gtype));
-    if (cinfo)
-        return cinfo;
-
-    if (!create_class)
-        return NULL;
-
-    c = Data_Make_Struct(rb_cData, RGObjClassInfo, cinfo_mark, NULL, cinfo);
+    data_type = rbgobj_class_info_create_data_type(klass);
+    c = TypedData_Make_Struct(rb_cObject, RGObjClassInfo, data_type, cinfo);
+    cinfo->klass = klass;
     cinfo->gtype = gtype;
     cinfo->mark  = NULL;
     cinfo->free  = NULL;
     cinfo->flags = 0;
+    cinfo->name = NULL;
+    cinfo->data_type = data_type;
+    rbgobj_class_info_fill_name(cinfo);
 
     fundamental_type = G_TYPE_FUNDAMENTAL(gtype);
-    switch (fundamental_type){
-    case G_TYPE_POINTER:
-    case G_TYPE_BOXED:
-    case G_TYPE_PARAM:
-    case G_TYPE_OBJECT:
-    case G_TYPE_ENUM:
-    case G_TYPE_FLAGS:
-        if (NIL_P(parent)) parent = get_superclass(gtype);
-        cinfo->klass = rb_funcall(rb_cClass, id_new, 1, parent);
-        break;
-
-    case G_TYPE_INTERFACE:
-        cinfo->klass = rb_module_new();
-        break;
-
-    default:
-      if (NIL_P(parent)) parent = get_superclass(gtype);
-      if (NIL_P(parent)) {
-          fprintf(stderr,
-                  "%s: %s's fundamental type %s isn't supported\n",
-                  "rbgobj_lookup_class_by_gtype",
-                  g_type_name(gtype),
-                  g_type_name(fundamental_type));
-          return NULL;
-      }
-      cinfo->klass = rb_funcall(rb_cClass, id_new, 1, parent);
-    }
-
     switch (fundamental_type){
       case G_TYPE_BOXED:
         rb_define_alloc_func(cinfo->klass, rbgobj_boxed_alloc_func);
@@ -217,10 +190,14 @@ rbgobj_lookup_class_by_gtype_without_lock(GType gtype, VALUE parent,
         interfaces = g_type_interfaces(gtype, &n_interfaces);
         for (i = 0; i < n_interfaces; i++){
             const RGObjClassInfo *iface_cinfo;
-            iface_cinfo =
-            rbgobj_lookup_class_by_gtype_without_lock(interfaces[i],
-                                                      Qnil,
-                                                      TRUE);
+            iface_cinfo = rbgobj_class_info_lookup_by_gtype(interfaces[i]);
+            if (!iface_cinfo) {
+                iface_cinfo =
+                    rbgobj_class_info_define_without_lock(interfaces[i],
+                                                          NULL,
+                                                          rb_cObject,
+                                                          Qnil);
+            }
             rb_include_module(cinfo->klass, iface_cinfo->klass);
         }
         g_free(interfaces);
@@ -253,55 +230,196 @@ rbgobj_lookup_class_by_gtype_without_lock(GType gtype, VALUE parent,
 }
 
 static VALUE
-rbgobj_lookup_class_by_gtype_body(VALUE data)
+get_superclass(GType gtype, VALUE module)
 {
-    RGObjClassByGtypeData *cdata = (RGObjClassByGtypeData *)data;
+    VALUE super_class;
+
+    if (rbgobj_convert_get_superclass(gtype, &super_class))
+        return super_class;
+
+    switch (gtype) {
+      case G_TYPE_PARAM:
+      case G_TYPE_OBJECT:
+        return cInstantiatable;
+      case G_TYPE_BOXED:
+      case G_TYPE_POINTER:
+      case G_TYPE_ENUM:
+      case G_TYPE_FLAGS:
+        return rb_cObject;
+      default:
+      {
+          GType parent_type;
+
+          parent_type = g_type_parent(gtype);
+          if (parent_type == G_TYPE_INVALID) {
+              return cInstantiatable;
+          } else {
+              const RGObjClassInfo *cinfo_super;
+              cinfo_super = rbgobj_class_info_lookup_by_gtype(parent_type);
+              if (!cinfo_super) {
+                  cinfo_super = rbgobj_class_info_define_without_lock(parent_type,
+                                                                      NULL,
+                                                                      module,
+                                                                      Qnil);
+              }
+              return cinfo_super->klass;
+          }
+      }
+    }
+}
+
+static RGObjClassInfo *
+rbgobj_class_info_define_without_lock(GType gtype,
+                                      const gchar *name,
+                                      VALUE module,
+                                      VALUE parent)
+{
+    GType fundamental_type;
+    VALUE klass;
+
+    fundamental_type = G_TYPE_FUNDAMENTAL(gtype);
+    switch (fundamental_type){
+    case G_TYPE_POINTER:
+    case G_TYPE_BOXED:
+    case G_TYPE_PARAM:
+    case G_TYPE_OBJECT:
+    case G_TYPE_ENUM:
+    case G_TYPE_FLAGS:
+        if (NIL_P(parent)) parent = get_superclass(gtype, module);
+        klass = rb_funcall(rb_cClass, id_new, 1, parent);
+        break;
+
+    case G_TYPE_INTERFACE:
+        klass = rb_module_new();
+        break;
+
+    default:
+      if (NIL_P(parent)) parent = get_superclass(gtype, module);
+      if (NIL_P(parent)) {
+          fprintf(stderr,
+                  "rbgobj_class_info_define: "
+                  "Unsupported fundamental type: <%s>(%s)\n",
+                  g_type_name(fundamental_type),
+                  g_type_name(gtype));
+          return NULL;
+      }
+      klass = rb_funcall(rb_cClass, id_new, 1, parent);
+    }
+
+    if (name) {
+        rb_define_const(module, name, klass);
+    }
+    return rbgobj_class_info_register_without_lock(gtype, klass);
+}
+
+static VALUE
+rbgobj_class_info_define_body(VALUE data)
+{
+    RGObjDefineClassData *cdata = (RGObjDefineClassData *)data;
     const RGObjClassInfo *cinfo;
 
-    cinfo = rbgobj_lookup_class_by_gtype_without_lock(cdata->gtype,
-                                                      cdata->parent,
-                                                      cdata->create_class);
+    cinfo = rbgobj_class_info_define_without_lock(cdata->gtype,
+                                                  cdata->name,
+                                                  cdata->module,
+                                                  cdata->parent);
     return (VALUE)cinfo;
 }
 
 static VALUE
-rbgobj_lookup_class_by_gtype_ensure(G_GNUC_UNUSED VALUE data)
+rbgobj_class_info_define_ensure(G_GNUC_UNUSED VALUE data)
 {
     rb_funcall(lookup_class_mutex, id_unlock, 0);
     return Qundef;
 }
 
+RGObjClassInfo *
+rbgobj_class_info_define(GType gtype,
+                         const gchar *name,
+                         VALUE module,
+                         VALUE parent)
+{
+    RGObjDefineClassData data;
+
+    data.gtype = gtype;
+    data.name = name;
+    data.module = module;
+    data.parent = parent;
+
+    rb_funcall(lookup_class_mutex, id_lock, 0);
+    return (RGObjClassInfo *)rb_ensure(rbgobj_class_info_define_body,
+                                       (VALUE)&data,
+                                       rbgobj_class_info_define_ensure,
+                                       (VALUE)&data);
+}
+
+/* deprecated */
+const RGObjClassInfo *
+rbgobj_lookup_class(VALUE klass)
+{
+    return rbgobj_class_info_lookup(klass);
+}
+
+const RGObjClassInfo *
+rbgobj_class_info_lookup(VALUE klass)
+{
+    VALUE data = rb_hash_aref(klass_to_cinfo, klass);
+    if (!NIL_P(data)) {
+        RGObjClassInfo *cinfo;
+        if (RTYPEDDATA_P(data)) {
+            TypedData_Get_Struct(data,
+                                 RGObjClassInfo,
+                                 RTYPEDDATA_TYPE(data),
+                                 cinfo);
+        } else {
+            Data_Get_Struct(data, RGObjClassInfo, cinfo);
+        }
+        return cinfo;
+    }
+
+    if (TYPE(klass) == T_CLASS) {
+        VALUE super;
+        super = rb_funcall(klass, id_superclass, 0);
+        return rbgobj_class_info_lookup(super);
+    }
+
+    rb_raise(rb_eRuntimeError, "can't get gobject class information");
+    return NULL;
+}
+
+const RGObjClassInfo *
+rbgobj_class_info_lookup_by_gtype(GType gtype)
+{
+    if (gtype == G_TYPE_INVALID)
+        return NULL;
+
+    return g_hash_table_lookup(gtype_to_cinfo, GUINT_TO_POINTER(gtype));
+}
+
+/* deprecated */
 const RGObjClassInfo *
 rbgobj_lookup_class_by_gtype(GType gtype, VALUE parent)
 {
     return rbgobj_lookup_class_by_gtype_full(gtype, parent, TRUE);
 }
 
+/* deprecated */
 const RGObjClassInfo *
 rbgobj_lookup_class_by_gtype_full(GType gtype, VALUE parent,
                                   gboolean create_class)
 {
     const RGObjClassInfo *info;
-    RGObjClassByGtypeData data;
 
-    info = rbgobj_lookup_class_by_gtype_without_lock(gtype, parent, FALSE);
-    if (info) {
-        return info;
-    }
-
-    if (!create_class) {
+    if (gtype == G_TYPE_INVALID)
         return NULL;
-    }
 
-    data.gtype = gtype;
-    data.parent = parent;
-    data.create_class = create_class;
+    info = rbgobj_class_info_lookup_by_gtype(gtype);
+    if (info)
+        return info;
 
-    rb_funcall(lookup_class_mutex, id_lock, 0);
-    return (RGObjClassInfo *)rb_ensure(rbgobj_lookup_class_by_gtype_body,
-                                       (VALUE)&data,
-                                       rbgobj_lookup_class_by_gtype_ensure,
-                                       (VALUE)&data);
+    if (!create_class)
+        return NULL;
+
+    return rbgobj_class_info_define(gtype, NULL, rb_cObject, parent);
 }
 
 VALUE
@@ -313,18 +431,49 @@ rbgobj_gtype_to_ruby_class(GType gtype)
     return cinfo ? cinfo->klass : Qnil;
 }
 
+GType
+rbgobj_gtype_from_ruby(VALUE rb_gtype)
+{
+    if (NIL_P(rb_gtype)) {
+        return G_TYPE_NONE;
+    } else if (RB_TYPE_P(rb_gtype, RUBY_T_STRING)) {
+        GType gtype;
+        gtype = g_type_from_name(RVAL2CSTR(rb_gtype));
+        if (gtype == G_TYPE_INVALID) {
+            rb_raise(rb_eArgError,
+                     "unknown GType name: <%s>",
+                     RVAL2CSTR(rb_gtype));
+        }
+        return gtype;
+    } else if (RVAL2CBOOL(rb_obj_is_kind_of(rb_gtype, RG_TARGET_NAMESPACE))) {
+        return NUM2SIZET(rb_ivar_get(rb_gtype, id_gtype));
+    } else if (RVAL2CBOOL(rb_obj_is_kind_of(rb_gtype, rb_cClass))) {
+        return CLASS2GTYPE(rb_gtype);
+    } else {
+        return NUM2SIZET(rb_to_int(rb_gtype));
+    }
+}
+
 VALUE
 rbgobj_define_class(GType gtype, const gchar *name, VALUE module,
                     RGMarkFunc mark, RGFreeFunc free, VALUE parent)
 {
-    RGObjClassInfo* cinfo;
-    if (gtype == 0)
-        rb_bug("rbgobj_define_class: Invalid gtype [%s]\n", name);
+    RGObjClassInfo *cinfo;
 
-    cinfo = (RGObjClassInfo*)rbgobj_lookup_class_by_gtype(gtype, parent);
+    if (gtype == G_TYPE_INVALID)
+        rb_bug("rbgobj_define_class: Invalid GType: <%s>\n", name);
+
+    cinfo = (RGObjClassInfo *)rbgobj_class_info_lookup_by_gtype(gtype);
+    if (cinfo) {
+        if (!rb_const_defined_at(module, rb_intern(name))) {
+            rb_define_const(module, name, cinfo->klass);
+            rbgobj_class_info_fill_name(cinfo);
+        }
+    } else {
+        cinfo = rbgobj_class_info_define(gtype, name, module, parent);
+    }
     cinfo->mark = mark;
     cinfo->free = free;
-    rb_define_const(module, name, cinfo->klass);
     return cinfo->klass;
 }
 
@@ -378,11 +527,14 @@ rbgobj_register_class(VALUE klass,
                       gboolean klass2gtype,
                       gboolean gtype2klass)
 {
-    RGObjClassInfo* cinfo = NULL;
+    rb_data_type_t *data_type = NULL;
+    RGObjClassInfo *cinfo = NULL;
     VALUE c = Qnil;
 
-    if (klass2gtype)
-        c = Data_Make_Struct(rb_cData, RGObjClassInfo, cinfo_mark, NULL, cinfo);
+    if (klass2gtype) {
+        data_type = rbgobj_class_info_create_data_type(klass);
+        c = TypedData_Make_Struct(rb_cObject, RGObjClassInfo, data_type, cinfo);
+    }
     if (gtype2klass && !cinfo)
         cinfo = g_new(RGObjClassInfo, 1);
 
@@ -392,6 +544,7 @@ rbgobj_register_class(VALUE klass,
         cinfo->mark  = NULL;
         cinfo->free  = NULL;
         cinfo->flags = 0;
+        cinfo->data_type = data_type;
     }
 
     if (klass2gtype)
@@ -455,26 +608,76 @@ init_typemap(void)
 /* GLib::Type */
 
 VALUE RG_TARGET_NAMESPACE;
-static ID id_gtype;
 
 VALUE
 rbgobj_gtype_new(GType gtype)
 {
     VALUE result = rb_obj_alloc(RG_TARGET_NAMESPACE);
-    VALUE arg = ULONG2NUM(gtype);
+    VALUE arg = SIZET2NUM(gtype);
     rb_obj_call_init(result, 1, &arg);
     return result;
 }
 
+/* Deprecated. Use rbgobj_gtype_from_ruby() instead. */
 GType
-rbgobj_gtype_get(VALUE self)
+rbgobj_gtype_get(VALUE rb_gtype)
 {
-    if (RVAL2CBOOL(rb_obj_is_kind_of(self, RG_TARGET_NAMESPACE))) {
-        return NUM2ULONG(rb_ivar_get(self, id_gtype));
-    } else {
-        return CLASS2GTYPE(self);
+    return rbgobj_gtype_from_ruby(rb_gtype);
+}
+
+static VALUE
+rg_s_try_convert(VALUE self, VALUE value)
+{
+    ID id_new;
+    CONST_ID(id_new, "new");
+
+    if (NIL_P(value))
+        return Qnil;
+
+    if (RVAL2CBOOL(rb_obj_is_kind_of(value, RG_TARGET_NAMESPACE)))
+        return value;
+
+    if (RVAL2CBOOL(rb_obj_is_kind_of(value, rb_cInteger))) {
+        GType gtype = NUM2ULONG(value);
+        if (!g_type_name(gtype))
+            return Qnil;
+        return rb_funcall(self, id_new, 1, value);
     }
-    rb_raise(rb_eTypeError, "Not a GLib::Type");
+
+    if (RB_TYPE_P(value, RUBY_T_SYMBOL)) {
+        value = rb_sym2str(value);
+    }
+
+    if (RB_TYPE_P(value, RUBY_T_STRING)) {
+        GType gtype = g_type_from_name(RVAL2CSTR(value));
+        if (gtype == G_TYPE_INVALID) {
+            return Qnil;
+        }
+        return rb_funcall(self, id_new, 1, value);
+    }
+
+    if (RVAL2CBOOL(rb_obj_is_kind_of(value, rb_cClass))) {
+        for (; !NIL_P(value); value = rb_funcall(value, id_superclass, 0)) {
+            RGObjClassInfo *cinfo;
+            VALUE data = rb_hash_aref(klass_to_cinfo, value);
+
+            if (NIL_P(data))
+                continue;
+
+            if (RTYPEDDATA_P(data)) {
+                TypedData_Get_Struct(data,
+                                     RGObjClassInfo,
+                                     RTYPEDDATA_TYPE(data),
+                                     cinfo);
+            } else {
+                Data_Get_Struct(data, RGObjClassInfo, cinfo);
+            }
+            return rb_funcall(self, id_new, 1, ULONG2NUM(cinfo->gtype));
+        }
+        return Qnil;
+    }
+
+    return Qnil;
 }
 
 static VALUE
@@ -483,7 +686,7 @@ rg_initialize(VALUE self, VALUE type)
     GType gtype;
 
     if (RVAL2CBOOL(rb_obj_is_kind_of(type, rb_cInteger))) {
-        gtype = NUM2ULONG(type);
+        gtype = NUM2SIZET(type);
         if (!g_type_name(gtype))
             gtype = G_TYPE_INVALID;
     } else {
@@ -493,7 +696,7 @@ rg_initialize(VALUE self, VALUE type)
     if (G_TYPE_INVALID == gtype)
         rb_raise(rb_eArgError, "invalid GType");
 
-    rb_ivar_set(self, id_gtype, ULONG2NUM(gtype));
+    rb_ivar_set(self, id_gtype, SIZET2NUM(gtype));
 
     return Qnil;
 }
@@ -501,7 +704,7 @@ rg_initialize(VALUE self, VALUE type)
 static VALUE
 rg_inspect(VALUE self)
 {
-    GType gtype = rbgobj_gtype_get(self);
+    GType gtype = rbgobj_gtype_from_ruby(self);
     gchar* str;
     VALUE result;
 
@@ -518,8 +721,8 @@ rg_operator_type_compare(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
 
         if (a==b)
             return INT2FIX(0);
@@ -538,8 +741,8 @@ rg_operator_type_eq(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(a == b);
     }
 }
@@ -550,8 +753,8 @@ rg_operator_type_lt_eq(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(g_type_is_a(a, b));
     }
 }
@@ -562,8 +765,8 @@ rg_operator_type_gt_eq(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(g_type_is_a(b, a));
     }
 }
@@ -574,8 +777,8 @@ rg_operator_type_lt(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(g_type_is_a(a, b) && a != b);
     }
 }
@@ -586,8 +789,8 @@ rg_operator_type_gt(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(g_type_is_a(b, a) && a != b);
     }
 }
@@ -601,112 +804,113 @@ rg_hash(VALUE self)
 static VALUE
 rg_to_class(VALUE self)
 {
-    return GTYPE2CLASS(rbgobj_gtype_get(self));
+    return GTYPE2CLASS(rbgobj_gtype_from_ruby(self));
 }
 
 static VALUE
 rg_fundamental(VALUE self)
 {
-    return rbgobj_gtype_new(G_TYPE_FUNDAMENTAL(rbgobj_gtype_get(self)));
+    return rbgobj_gtype_new(G_TYPE_FUNDAMENTAL(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_fundamental_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_FUNDAMENTAL(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_FUNDAMENTAL(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_derived_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_DERIVED(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_DERIVED(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_interface_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_INTERFACE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_INTERFACE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_classed_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_CLASSED(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_CLASSED(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_instantiatable_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_INSTANTIATABLE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_INSTANTIATABLE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_derivable_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_DERIVABLE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_DERIVABLE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_deep_derivable_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_DEEP_DERIVABLE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_DEEP_DERIVABLE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_abstract_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_ABSTRACT(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_ABSTRACT(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_value_abstract_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_VALUE_ABSTRACT(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_VALUE_ABSTRACT(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_value_type_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_VALUE_TYPE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_VALUE_TYPE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_has_value_table(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_HAS_VALUE_TABLE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_HAS_VALUE_TABLE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_name(VALUE self)
 {
-    return rb_str_new2(g_type_name(rbgobj_gtype_get(self)));
+    return rb_str_new2(g_type_name(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_parent(VALUE self)
 {
-    GType parent = g_type_parent(rbgobj_gtype_get(self));
+    GType parent = g_type_parent(rbgobj_gtype_from_ruby(self));
     return parent ? rbgobj_gtype_new(parent) : Qnil;
 }
 
 static VALUE
 rg_depth(VALUE self)
 {
-    return UINT2NUM(g_type_depth(rbgobj_gtype_get(self)));
+    return UINT2NUM(g_type_depth(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_next_base(VALUE leaf_type, VALUE root_type)
 {
-    GType ret = g_type_next_base(rbgobj_gtype_get(leaf_type),
-                                 rbgobj_gtype_get(root_type));
+    GType ret = g_type_next_base(rbgobj_gtype_from_ruby(leaf_type),
+                                 rbgobj_gtype_from_ruby(root_type));
     return ret ? rbgobj_gtype_new(ret) : Qnil;
 }
 
 static VALUE
 rg_type_is_a_p(VALUE self, VALUE is_a_type)
 {
-    return CBOOL2RVAL(g_type_is_a(rbgobj_gtype_get(self), rbgobj_gtype_get(is_a_type)));
+    return CBOOL2RVAL(g_type_is_a(rbgobj_gtype_from_ruby(self),
+                                  rbgobj_gtype_from_ruby(is_a_type)));
 }
 
 #if 0
@@ -727,7 +931,7 @@ rg_children(VALUE self)
     VALUE result;
     guint i;
 
-    types = g_type_children(rbgobj_gtype_get(self), &n_children);
+    types = g_type_children(rbgobj_gtype_from_ruby(self), &n_children);
     result = rb_ary_new2(n_children);
     for (i = 0; i < n_children; i++)
         rb_ary_store(result, i, rbgobj_gtype_new(types[i]));
@@ -744,7 +948,7 @@ rg_interfaces(VALUE self)
     VALUE result;
     guint i;
 
-    types = g_type_interfaces(rbgobj_gtype_get(self), &n_interfaces);
+    types = g_type_interfaces(rbgobj_gtype_from_ruby(self), &n_interfaces);
     result = rb_ary_new2(n_interfaces);
     for (i = 0; i < n_interfaces; i++)
         rb_ary_store(result, i, rbgobj_gtype_new(types[i]));
@@ -757,7 +961,7 @@ static VALUE
 rg_class_size(VALUE self)
 {
     GTypeQuery query;
-    g_type_query(rbgobj_gtype_get(self), &query);
+    g_type_query(rbgobj_gtype_from_ruby(self), &query);
     return UINT2NUM(query.class_size);
 }
 
@@ -765,7 +969,7 @@ static VALUE
 rg_instance_size(VALUE self)
 {
     GTypeQuery query;
-    g_type_query(rbgobj_gtype_get(self), &query);
+    g_type_query(rbgobj_gtype_from_ruby(self), &query);
     return UINT2NUM(query.instance_size);
 }
 
@@ -773,6 +977,7 @@ static inline void
 _def_fundamental_type(VALUE ary, GType gtype, const char* name)
 {
     VALUE c = rbgobj_gtype_new(gtype);
+    rb_obj_freeze(c);
     rb_define_const(RG_TARGET_NAMESPACE, name, c);
     rb_ary_push(ary, c);
 }
@@ -782,10 +987,6 @@ _def_fundamental_type(VALUE ary, GType gtype, const char* name)
 void
 Init_gobject_gtype(void)
 {
-#if !GLIB_CHECK_VERSION(2, 35, 1)
-    g_type_init();
-#endif
-
     init_typemap();
 
     /* type */
@@ -795,12 +996,14 @@ Init_gobject_gtype(void)
     id_lock = rb_intern("lock");
     id_unlock = rb_intern("unlock");
     lookup_class_mutex = rb_funcall(rb_cMutex, id_new, 0);
-    rb_iv_set(mGLib, "lookup_class_mutex", lookup_class_mutex);
+    rb_iv_set(rbg_mGLib(), "lookup_class_mutex", lookup_class_mutex);
 
     dynamic_gtype_list = g_hash_table_new(g_str_hash, g_str_equal);
     id_gtype = rb_intern("__gobject_gtype__");
 
-    RG_TARGET_NAMESPACE = rb_define_class_under(mGLib, "Type", rb_cObject);
+    RG_TARGET_NAMESPACE = rb_define_class_under(rbg_mGLib(), "Type", rb_cObject);
+
+    RG_DEF_SMETHOD(try_convert, 1);
 
     rb_define_alias(CLASS_OF(RG_TARGET_NAMESPACE), "[]", "new");
     RG_DEF_METHOD(initialize, 1);
